""" Shared tests for all Tags functions."""

import time

import pytest
from conftest import make_random_image, make_random_tabular, make_random_text
from nyckel import (
    ClassificationLabel,
    ClassificationPrediction,
    ImageTagsFunction,
    ImageTagsSample,
    TabularFunctionField,
    TabularTagsFunction,
    TabularTagsSample,
    TagsAnnotation,
    TextTagsFunction,
    TextTagsSample,
)


@pytest.mark.parametrize(
    "function_type,function_class",
    [
        ("image_tags_function", ImageTagsFunction),
        ("text_tags_function", TextTagsFunction),
        ("tabular_tags_function", TabularTagsFunction),
    ],
)
class TestFunction:

    # Create & Delete happens in the test fixture, and there is no update operation
    # So remains to test read.

    def test_read(self, function_type, function_class, auth_test_credentials, request):
        func = request.getfixturevalue(function_type)
        function_id = func.function_id
        reloaded_function = function_class(function_id, auth_test_credentials)
        assert reloaded_function.function_id == function_id


@pytest.mark.parametrize(
    "function_class,sample_data_maker",
    [
        (ImageTagsFunction, make_random_image),
        (TextTagsFunction, make_random_text),
        (TabularTagsFunction, make_random_tabular),
    ],
)
class TestProperties:

    def test_all(self, function_class, sample_data_maker, auth_test_credentials):
        name = "PYTHON-SDK TAGS PROPERTIES TEST FUNCTION"
        func = function_class.create(name, auth_test_credentials)
        if isinstance(func, TabularTagsFunction):
            fields = [
                TabularFunctionField(name="name", type="Text"),
                TabularFunctionField(name="age", type="Number"),
                TabularFunctionField(name="mug", type="Image"),
            ]
            func.create_fields(fields)
        assert func.name == name
        assert func.sample_count == 0
        assert func.label_count == 0

        func.create_labels(["cat", "dog"])
        time.sleep(0.5)
        assert func.label_count == 2

        func.create_samples([sample_data_maker(), sample_data_maker()])
        time.sleep(0.5)
        assert func.sample_count == 2

        func.delete()


@pytest.mark.parametrize("function_type", ["image_tags_function", "text_tags_function", "tabular_tags_function"])
class TestLabels:

    def test_create_untyped(self, function_type, request):
        func = request.getfixturevalue(function_type)
        label_ids = func.create_labels(["cat", "dog"])
        assert len(label_ids) == 2

    def test_create_typed(self, function_type, request):
        func = request.getfixturevalue(function_type)
        label_ids = func.create_labels([ClassificationLabel("cat"), ClassificationLabel("dog")])
        assert len(label_ids) == 2

    def test_read(self, function_type, request):
        func = request.getfixturevalue(function_type)
        label_id = func.create_labels(["dog"])[0]
        label = func.read_label(label_id)
        assert label.name == "dog"
        assert label.id == label_id

    def test_update(self, function_type, request):
        func = request.getfixturevalue(function_type)
        label_id = func.create_labels(["dog"])[0]
        updated_label = ClassificationLabel(name="doggo", id=label_id)
        func.update_label(updated_label)
        time.sleep(0.5)
        labels = func.list_labels()
        assert labels[0].name == "doggo"

    def test_delete(self, function_type, request):
        func = request.getfixturevalue(function_type)
        label_id = func.create_labels(["dog"])[0]
        func.delete_labels([label_id])
        time.sleep(0.5)
        labels = func.list_labels()
        assert len(labels) == 0

    def test_strip_whitespaces(self, function_type, request):
        func = request.getfixturevalue(function_type)
        func.create_labels(["  cat  "])
        labels = func.list_labels()
        assert labels[0].name == "cat"


@pytest.mark.parametrize(
    "function_type,sample_data_maker,sample_class",
    [
        ("image_tags_function", make_random_image, ImageTagsSample),
        ("text_tags_function", make_random_text, TextTagsSample),
        ("tabular_tags_function_with_fields", make_random_tabular, TabularTagsSample),
    ],
)
class TestSamples:

    def test_create_untyped(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)
        sample_ids = func.create_samples([sample_data_maker(), sample_data_maker()])
        assert len(sample_ids) == 2

    def test_create_typed(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)
        sample_ids = func.create_samples(
            [sample_class(data=sample_data_maker()), sample_class(data=sample_data_maker())]
        )
        assert len(sample_ids) == 2

    def test_create_annotated(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)
        sample_ids = func.create_samples([sample_class(data=sample_data_maker(), annotation=[TagsAnnotation("Nice")])])
        assert len(sample_ids) == 1
        assert func.label_count == 1  # Labels are created automatically

    def test_read(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)
        sample_to_be_created = sample_data_maker()
        sample_id = func.create_samples([sample_to_be_created])[0]
        time.sleep(0.5)
        sample = func.read_sample(sample_id)
        assert sample.id == sample_id
        if function_type == "text_tags_function":
            # We can't test for exact equality for image and (tabular with image fields) because
            # the image data is recoded by the server.
            assert sample.data == sample_to_be_created

    def test_update(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)
        func.create_labels(["Nice"])
        original_sample = sample_class(data=sample_data_maker())
        sample_id = func.create_samples([original_sample])[0]
        time.sleep(0.5)

        updated_sample = sample_class(id=sample_id, data=original_sample.data, annotation=[TagsAnnotation("Nice")])
        func.update_annotation(updated_sample)
        time.sleep(0.5)

        sample = func.read_sample(sample_id)
        assert sample.annotation[0] == TagsAnnotation("Nice")

    def test_update_with_two_samples(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)

        samples = [
            sample_class(data=sample_data_maker(), annotation=[TagsAnnotation(label_name="Nice", present=True)]),
            sample_class(data=sample_data_maker(), annotation=[TagsAnnotation(label_name="Bad", present=True)]),
        ]

        func.create_samples(samples)
        time.sleep(0.5)

        samples = func.list_samples()  # type: ignore
        samples[0].annotation = [
            TagsAnnotation(label_name="Nice", present=True),
            TagsAnnotation(label_name="Bad", present=True),
        ]
        func.update_annotation(samples[0])
        time.sleep(0.5)

        assert samples[0].id is not None
        updated_sample = func.read_sample(samples[0].id)
        assert updated_sample.annotation == samples[0].annotation

    def test_delete(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)
        sample_id = func.create_samples([sample_data_maker()])[0]
        func.delete_samples([sample_id])
        time.sleep(0.5)
        samples = func.list_samples()
        assert len(samples) == 0

    def test_new_labels(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)

        func.create_labels([ClassificationLabel(name="label1")])
        sample_1_id = func.create_samples(
            [sample_class(data=sample_data_maker(), annotation=[TagsAnnotation("label1")])]
        )[0]

        func.create_labels([ClassificationLabel(name="label2")])

        sample_2_id = func.create_samples(
            [sample_class(data=sample_data_maker(), annotation=[TagsAnnotation("label1")])]
        )[0]
        time.sleep(0.5)

        sample1 = func.read_sample(sample_1_id)

        # When sample1 was created, label2 wasn't present yet.
        # So sample1 doesn't have an "opinion" about that label.
        # This means that it's not present in the annotation field.
        assert len(sample1.annotation) == 1
        assert TagsAnnotation("label1") in sample1.annotation
        assert TagsAnnotation("label2", present=False) not in sample1.annotation

        sample2 = func.read_sample(sample_2_id)
        # When sample2 was created, both labels were present.
        # Since we didn't give an annotation for label2, it was set to not present.
        assert len(sample2.annotation) == 2
        assert TagsAnnotation("label1") in sample2.annotation
        assert TagsAnnotation("label2", present=False) in sample2.annotation


@pytest.mark.parametrize(
    "function_type,sample_data_maker,sample_class",
    [
        ("image_tags_function", make_random_image, ImageTagsSample),
        ("text_tags_function", make_random_text, TextTagsSample),
        ("tabular_tags_function_with_fields", make_random_tabular, TabularTagsSample),
    ],
)
class TestEndToEnd:

    def test_end_to_end(self, function_type, sample_data_maker, sample_class, request):
        func = request.getfixturevalue(function_type)

        labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
        func.create_labels(labels_to_create)

        samples = [
            sample_class(data=sample_data_maker(), annotation=[TagsAnnotation("Nice")]),
            sample_class(data=sample_data_maker(), annotation=[TagsAnnotation("Nice"), TagsAnnotation("Boo")]),
            sample_class(data=sample_data_maker(), annotation=[TagsAnnotation("Boo")]),
            sample_class(data=sample_data_maker(), annotation=[TagsAnnotation("Boo")]),
            sample_class(data=sample_data_maker(), annotation=[]),
        ]
        assert not func.has_trained_model()
        func.create_samples(samples)

        while not func.has_trained_model():
            print("-> No trained model yet. Sleeping 1 sec...")
            time.sleep(1)

        predictions = func.invoke([sample_data_maker(), sample_data_maker(), sample_data_maker()])
        assert len(predictions) == 3
        for prediction in predictions:
            if len(prediction) > 0:
                assert isinstance(prediction[0], ClassificationPrediction)

        returned_samples = func.list_samples()
        assert len(samples) == len(returned_samples)
        assert isinstance(returned_samples[0], sample_class)



import time
from typing import Callable, Dict, List, Sequence, Tuple, Union

from tqdm import tqdm

from nyckel import (
    ClassificationAnnotation,
    ClassificationFunction,
    ClassificationLabel,
    ClassificationPrediction,
    Credentials,
    LabelName,
    NyckelId,
    TabularClassificationSample,
    TabularFunctionField,
    TabularSampleData,
)
from nyckel.functions.classification import factory
from nyckel.functions.classification.classification import ClassificationFunctionURLHandler
from nyckel.functions.classification.function_handler import ClassificationFunctionHandler
from nyckel.functions.classification.label_handler import ClassificationLabelHandler
from nyckel.functions.classification.sample_handler import ClassificationSampleHandler
from nyckel.functions.utils import ImageFieldTransformer, strip_nyckel_prefix
from nyckel.request_utils import ParallelPoster, SequentialGetter


class TabularClassificationFunction(ClassificationFunction):
    """
    Example:

    ```py

    from nyckel import Credentials, TabularClassificationFunction, TabularFunctionField

    credentials = Credentials(client_id="...", client_secret="...")

    func = TabularClassificationFunction.create("InterestedProspect", credentials)
    func.create_fields([
        TabularFunctionField(type="Text", name="Name"),
        TabularFunctionField(type="Text", name="Response")
    ])
    func.create_samples([
        ({"Name": "Adam Adams", "Response": "Thanks for reaching out. I'd love to chat"}, "Interested"),
        ({"Name": "Bo Berg", "Response": "Sure! Can you tell me a bit more?"}, "Interested"),
        ({"Name": "Charles Carter", "Response": "No thanks, I don't need a Classification API"}, "Not Interested"),
        ({"Name": "Devin Duncan", "Response": "Nope. Please stop bugging me."}, "Not Interested"),
    ])

    predictions = func.invoke([{"Name": "Frank Fisher", "Response": "Yea, I'd love to try the Nyckel API!"}])
    ```
    """

    def __init__(self, function_id: NyckelId, credentials: Credentials) -> None:
        self._function_id = function_id

        self._function_handler = ClassificationFunctionHandler(function_id, credentials)
        self._label_handler = ClassificationLabelHandler(function_id, credentials)
        self._field_handler = TabularFieldHandler(function_id, credentials)
        self._sample_handler = ClassificationSampleHandler(function_id, credentials)
        self._url_handler = ClassificationFunctionURLHandler(function_id, credentials.server_url)
        assert self._function_handler.get_input_modality() == "Tabular"

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        status_string = f"Name: {self.name}, id: {self.function_id}, url: {self._url_handler.train_page}"
        return status_string

    @property
    def function_id(self) -> str:
        return self._function_id

    @property
    def sample_count(self) -> int:
        return self._function_handler.sample_count

    @property
    def label_count(self) -> int:
        return self._function_handler.label_count

    @property
    def name(self) -> str:
        return self._function_handler.get_name()

    @classmethod
    def create(cls, name: str, credentials: Credentials) -> "TabularClassificationFunction":
        return factory.ClassificationFunctionFactory.create(name, "Tabular", credentials)  # type: ignore

    def delete(self) -> None:
        self._function_handler.delete()

    def invoke(  # type: ignore
        self,
        sample_data_list: List[TabularSampleData],
        model_id: str = "",
    ) -> List[ClassificationPrediction]:
        return self._sample_handler.invoke(sample_data_list, self._get_image_field_transformer(), model_id=model_id)

    def has_trained_model(self) -> bool:
        return self._function_handler.is_trained

    def create_labels(self, labels: Sequence[Union[ClassificationLabel, str]]) -> List[NyckelId]:
        typed_labels = [
            label if isinstance(label, ClassificationLabel) else ClassificationLabel(name=label) for label in labels
        ]
        return self._label_handler.create_labels(typed_labels)

    def list_labels(self) -> List[ClassificationLabel]:
        return self._label_handler.list_labels(self.label_count)

    def read_label(self, label_id: NyckelId) -> ClassificationLabel:
        return self._label_handler.read_label(label_id)

    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        return self._label_handler.update_label(label)

    def delete_labels(self, label_ids: List[NyckelId]) -> None:
        return self._label_handler.delete_labels(label_ids)

    def create_fields(self, fields: List[TabularFunctionField]) -> List[NyckelId]:
        return self._field_handler.create_fields(fields)

    def list_fields(self) -> List[TabularFunctionField]:
        return self._field_handler.list_fields()

    def read_field(self, field_id: NyckelId) -> TabularFunctionField:
        return self._field_handler.read_field(field_id)

    def delete_field(self, field_id: NyckelId) -> None:
        return self._field_handler.delete_field(field_id)

    def create_samples(self, samples: Sequence[Union[TabularClassificationSample, Tuple[TabularSampleData, LabelName], TabularSampleData]]) -> List[NyckelId]:  # type: ignore # noqa: E501
        if len(samples) == 0:
            return []

        typed_samples = self._wrangle_post_samples_input(samples)
        typed_samples = self._strip_label_names(typed_samples)
        self._assert_fields_created(typed_samples)
        self._create_labels_as_needed(typed_samples)
        return self._sample_handler.create_samples(typed_samples, self._get_image_field_transformer())

    def _get_image_field_transformer(self) -> Callable:
        fields = self.list_fields()
        image_field_transformer = lambda x: x  # noqa: E731
        for field in fields:
            if field.type == "Image":
                # There is only one image field (max) per function, so we can break here.
                image_field_transformer = ImageFieldTransformer(field.name)
                break
        return image_field_transformer

    def list_samples(self) -> List[TabularClassificationSample]:  # type: ignore
        samples_dict_list = self._sample_handler.list_samples(self.sample_count)
        labels = self._label_handler.list_labels(None)
        fields = self.list_fields()

        label_name_by_id = {label.id: label.name for label in labels}
        field_name_by_id = {field.id: field.name for field in fields}  # type: ignore

        return [self._sample_from_dict(entry, label_name_by_id, field_name_by_id) for entry in samples_dict_list]  # type: ignore # noqa: E501

    def read_sample(self, sample_id: NyckelId) -> TabularClassificationSample:
        sample_as_dict = self._sample_handler.read_sample(sample_id)

        labels = self._label_handler.list_labels(None)
        fields = self.list_fields()

        label_name_by_id = {label.id: label.name for label in labels}
        field_name_by_id = {field.id: field.name for field in fields}  # type: ignore

        return self._sample_from_dict(sample_as_dict, label_name_by_id, field_name_by_id)  # type: ignore

    def update_annotation(self, sample: TabularClassificationSample) -> None:  # type: ignore
        self._sample_handler.update_annotation(sample)

    def delete_samples(self, sample_ids: List[NyckelId]) -> None:
        self._sample_handler.delete_samples(sample_ids)

    def _wrangle_post_samples_input(
        self,
        samples: Sequence[Union[TabularClassificationSample, Tuple[TabularSampleData, LabelName], TabularSampleData]],
    ) -> List[TabularClassificationSample]:
        typed_samples: List[TabularClassificationSample] = []
        for sample in samples:
            if isinstance(sample, TabularClassificationSample):
                typed_samples.append(sample)
            elif isinstance(sample, (list, tuple)):
                data_dict, label_name = sample
                typed_samples.append(
                    TabularClassificationSample(
                        data=data_dict, annotation=ClassificationAnnotation(label_name=label_name)
                    )
                )
            elif isinstance(sample, dict):
                typed_samples.append(TabularClassificationSample(data=sample))
            else:
                raise ValueError(f"Unknown sample type: {type(sample)}")
        return typed_samples

    def _assert_fields_created(self, samples: List[TabularClassificationSample]) -> None:
        existing_fields = self.list_fields()
        existing_field_names = {field.name for field in existing_fields}
        new_field_names = {field_name for sample in samples for field_name in sample.data.keys()}
        missing_field_names = new_field_names - existing_field_names
        assert len(missing_field_names) == 0, f"Fields not created: {missing_field_names=}. Please create fields first."

    def _create_labels_as_needed(self, samples: List[TabularClassificationSample]) -> None:
        existing_labels = self._label_handler.list_labels(None)
        existing_label_names = {label.name for label in existing_labels}
        new_label_names = {sample.annotation.label_name for sample in samples if sample.annotation}
        missing_label_names = new_label_names - existing_label_names
        missing_labels = [ClassificationLabel(name=label_name) for label_name in missing_label_names]
        if len(missing_labels) > 0:
            self._label_handler.create_labels(missing_labels)

    def _sample_from_dict(
        self, sample_dict: Dict, label_name_by_id: Dict[str, str], field_name_by_id: Dict[str, str]
    ) -> TabularClassificationSample:
        tabular_data_body = {
            field_name_by_id[strip_nyckel_prefix(field_id)]: field_data
            for field_id, field_data in sample_dict["data"].items()
        }

        if "externalId" in sample_dict:
            external_id = sample_dict["externalId"]
        else:
            external_id = None

        if "annotation" in sample_dict:
            annotation = ClassificationAnnotation(
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["annotation"]["labelId"])],
            )
        else:
            annotation = None

        if "prediction" in sample_dict:
            prediction = ClassificationPrediction(
                confidence=sample_dict["prediction"]["confidence"],
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["prediction"]["labelId"])],
            )
        else:
            prediction = None

        return TabularClassificationSample(
            id=strip_nyckel_prefix(sample_dict["id"]),
            data=tabular_data_body,
            external_id=external_id,
            annotation=annotation,
            prediction=prediction,
        )

    def _strip_label_names(self, samples: List[TabularClassificationSample]) -> List[TabularClassificationSample]:
        for sample in samples:
            if sample.annotation:
                sample.annotation.label_name = sample.annotation.label_name.strip()
        return samples


class TabularFieldHandler:
    def __init__(self, function_id: NyckelId, credentials: Credentials):
        self._function_id = function_id
        self._credentials = credentials
        self._url_handler = ClassificationFunctionURLHandler(function_id, credentials.server_url)
        self._field_name_by_id: Dict = {}

    def create_fields(self, fields: List[TabularFunctionField]) -> List[NyckelId]:
        bodies = [{"name": field.name, "type": field.type} for field in fields]
        url = self._url_handler.api_endpoint(path="fields")
        session = self._credentials.get_session()
        progress_bar = tqdm(total=len(bodies), ncols=80, desc="Posting fields")
        responses = ParallelPoster(session, url, progress_bar)(bodies)
        field_ids = [strip_nyckel_prefix(resp.json()["id"]) for resp in responses]

        self._confirm_new_fields_available(fields)
        return field_ids

    def _confirm_new_fields_available(self, new_fields: List[TabularFunctionField]) -> None:
        # Before returning, make sure the assets are available via the API.
        new_fields_available_via_api = False
        timeout_seconds = 5
        t0 = time.time()
        while not new_fields_available_via_api:
            if time.time() - t0 > timeout_seconds:
                raise ValueError("Something went wrong when posting fields.")
            time.sleep(0.5)
            fields_retrieved = self.list_fields()
            new_fields_available_via_api = set([label.name for label in new_fields]).issubset(
                [label.name for label in fields_retrieved]
            )

    def list_fields(self) -> List[TabularFunctionField]:
        session = self._credentials.get_session()
        fields_dict_list = SequentialGetter(session, self._url_handler.api_endpoint(path="fields"))(
            tqdm(ncols=80, desc="Listing fields")
        )
        return [self._field_from_dict(entry) for entry in fields_dict_list]

    def read_field(self, field_id: NyckelId) -> TabularFunctionField:
        session = self._credentials.get_session()
        url = self._url_handler.api_endpoint(path=f"fields/{field_id}")
        response = session.get(url)
        if not response.status_code == 200:
            raise RuntimeError(
                f"Unable to fetch field {field_id} from {self._url_handler.train_page} "
                f"{response.text=} {response.status_code=}"
            )
        return self._field_from_dict(response.json())

    def delete_field(self, field_id: NyckelId) -> None:
        session = self._credentials.get_session()
        response = session.delete(self._url_handler.api_endpoint(path=f"fields/{field_id}"))
        assert response.status_code == 200, f"Delete failed with {response.status_code=}, {response.text=}"

    def _field_from_dict(self, field_dict: Dict) -> TabularFunctionField:
        return TabularFunctionField(
            name=field_dict["name"], id=strip_nyckel_prefix(field_dict["id"]), type=field_dict["type"]
        )



import os

from conftest import large_image_url, make_random_image, small_image_url
from nyckel import ImageTagsFunction
from PIL import Image


class TestSamples:

    def test_create_inline(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        sample_ids = func.create_samples([make_random_image(), make_random_image()])
        assert len(sample_ids) == 2

    def test_create_PIL(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        sample_ids = func.create_samples([Image.new("RGB", (100, 100)), Image.new("RGB", (80, 80))])
        assert len(sample_ids) == 2

    def test_create_local_path(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        local_filepath = os.path.abspath("tests/fixtures/flower.jpg")
        sample_ids = func.create_samples([local_filepath])
        assert len(sample_ids) == 1

    def test_create_url(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        sample_ids = func.create_samples([small_image_url, large_image_url])
        assert len(sample_ids) == 2
